import logging,json,simplejson
#from posixpath import split
from . import restClient as client
from . import Sobjects,utils
" select fields(all) from vlocity_cmt__EntityFilter__c where Id in ('a5W7Z0000009gEiUAI','a5W7Z0000009gdtUAA','a5W7Z0000009gFfUAI','a5W7Z0000009gEdUAI')"

def _query(q,raiseEx=True,logResponse = False):
    logging.debug(q)
    call = client.callAPI(f'/services/data/v55.0/query?q={q}')
    lc = client.lastThreadCall()

    lc['query']=q
    if lc['error'] != None:
        lc['query'] = q
        if raiseEx==True:
            utils.raiseException(lc['errorCode'],lc['error'],'query',f"{lc['errorOther']} . {q}")
            raise ValueError(simplejson.dumps(call, indent=4))
        logging.warn(simplejson.dumps(call, indent=4))

        return None

    if logResponse == True:
        utils.printJSON(call)

    return call

#--------------------------------------------------------------------------------------------------------------------------------------------------------
def query(q,raiseEx=True,X=['Id','Name'],logResponse=False):
    """
    Executes a query in Salesforce.
    - q: the query string. "select.... from... limit..."
        the where clause can specify a normal field or $X and $Z. 
    - X: by default it will query for Id and Name. 
    - logResponse: 
    """
    if '$X=' not in q and '$Z=' not in q:
        return _query(q,raiseEx,logResponse)

    if '$Z='  in q:
        X=['Id','ProductCode','Name']
        q = q.replace('$Z=','$X=')

    selectfields = q.strip().split()

    for x in range(len(selectfields)):
        if selectfields[x].lower() == 'from':
            objectType = selectfields[x+1]

        if  '$X' in selectfields[x]:
            where = selectfields[x].split("'")
            value = where[1]

    for field in X:
        if field == 'Id':
            if Sobjects.isId(objectType,value) == False:
                continue
        q1 = q.replace('$X',field)
        call = _query(q1,raiseEx=False)      
        if call['totalSize'] > 0:
            return call

    return None

def queryRecords(q,raiseEx=True,X=['Id','Name']):
    select = query(q,raiseEx,X)
    if select == None:
        return None
    return select['records']

cache = {}
def queryFieldList(q,field=None,raiseEx=False,X=['Id','Name']):
    if '$X=' in q or '$Z=' in q:
        print(q)
        if q in cache:
            return cache[q]
    records = queryRecords(q,raiseEx,X)

    if records == None:
        return None

    if field == None:
        field = q.strip().split()[1]
    
    fieldList = []
  
    for record in records:
        fieldList.append(record[field])
        
    if '$X=' in q or '$Z=' in q:
        cache[q]=fieldList

    return fieldList

#" select Id from vlocity_cmt__AttributeAssignment__c where vlocity_cmt__AttributeUniqueCode__c='ATT_MOBILE_CREDITS' and  vlocity_cmt__ObjectId__c='01t7Z00000AVpCLQA1' "
def queryField(q,field=None,raiseEx=False,X=['Id','Name']):
    fieldList = queryFieldList(q,field,raiseEx,X)

    if fieldList == None or len(fieldList) == 0:
        return None

    if len(fieldList) > 1:
        logging.warn(f"There is more than one record returned. total records {len(fieldList)}, query={q}")

    return fieldList[0]

def queryIdF(objName,extendedF):
    if extendedF == None:
        return None
    ef = utils.extendedField(extendedF)
    if ef['field'] == 'Id':
        return ef['value']
    return queryField(f" select Id from {objName} where {ef['field']} = '{ef['value']}' ")
def logQuery(q,filename,raiseEx=False):
    if filename != None:
        client.logCall(filename)
    call = query(q,raiseEx)

    return call


def process( call):
    for asset in call['records']:
        asset['vlocity_cmt__PricingLogData__c'] = json.loads( asset['vlocity_cmt__PricingLogData__c'])

def take_snapshot(fileName,q,postProcessing=None):
    client.debugFile(fileName)
    client.setDebugReplyProcessing(postProcessing)
    query(q)

def records(call):
    return call['records']

def IN_clause(list):
    values = []
    for p in list:
        values.append(f"'{p}'")
    return ",".join(values)


import unittest
from InCli import InCli
from InCli.SFAPI import file

class Test_File(unittest.TestCase):
    def test_main(self):
        filename = "testFile"
        j1 = "This is a sentence."
        j1n = "This is a sentence.\n"

        filepath = file.write(filename,j1)

        j2 = file.read(filename)
        self.assertTrue(j1==j2)

        j3 = file.read(filepath)
        self.assertTrue(j1==j3)

        file.delete(filename)

        self.assertTrue(file.exists(filename)==False)

        f = file.openFile(filename)
        file.write_line(f,j1)
        file.closeFile(f)

        j4 = file.read(filename)
        self.assertTrue(j1n==j4)

        print()

        



import logging
from . import utils
from . import restClient as client
from . import query
import re,simplejson

apiVersion = 'v51.0'

#X can be an Id or the field value. If it is an Id, it is returned. If not, X it the value of the 'field'. 

def getFieldList(objs,name):
    """
    From and array of objects, return an array of field by name. Return a column. 
    """
    return [obj['value'] for obj in objs]
    fields = []
    for obj in objs:
        fields.append(obj[name])

    return fields

def Id(obj):
    if type(obj) is str:
        return obj
    if type(obj) is dict:
        if 'Id' in obj:
            id = obj['Id']
        if 'id' in obj:
            id = obj['id']
        if 'value' in id:
            id = id['value']
        return id

def checkId(id):
    #return re.search(r"[a-zA-Z0-9]{15}|[a-zA-Z0-9]{18}", id)
    return re.search(r"\b[a-z0-9]\w{4}0\w{12}|[a-z0-9]\w{4}0\w{9}\b", id)


def isId(sobjectType,X):
    if X == None:
        return False
    if type(X) is dict:
        return isId(utils.Id(X))
    if sobjectType.lower() == str(getSObjectType(X)).lower():
        return True   
    return False

#---------------------------------
def getSObjectType(id):
    if checkId(id) == None:
        return None
    if id.startswith("a3O"):
        return "vlocity_cmt__PriceList__c"
    if id.startswith("01s"):
        return "Pricebook2"
    if id.startswith("01t"):
        return "product2"
    if id.startswith("a3P"):
        return "vlocity_cmt__PricingElement__c"
    if id.startswith("a4d"):
        return "vlocity_cmt__PricingElement__c"
    if id.startswith("a36"):
        return "vlocity_cmt__Promotion__c"
    if id.startswith("a3S"):
        return "vlocity_cmt__PromotionItem__c"
    if id.startswith("a3i"):
        return "vlocity_cmt__PriceListEntry__c"
    if id.startswith("a3R"):
        return "vlocity_cmt__PricingVariable__c"
    if id.startswith("a4h"):
        return "vlocity_cmt__PricingVariable__c"
    if id.startswith("a1b"):
        return "vlocity_cmt__ProductChildItem__c"
    if id.startswith("a2S"):
        return "vlocity_cmt__ObjectClass__c"
    if id.startswith("a4M"):
        return "vlocity_cmt__PricingPlan__c"
    if id.startswith("a3T"):
        return "vlocity_cmt__TimePlan__c"
    if id.startswith("a3l"):
        return "vlocity_cmt__TimePolicy__c"
    if id.startswith("001"):
        return "Account"     
    if id.startswith("801"):
        return "Order"    
    if id.startswith("802"):
        return "OrderItem"    
    if id.startswith("02i"):
        return "Asset"    
    if id.startswith("01s"):
        return "Pricebook2" 
    if id.startswith("a0I"):
        return "vlocity_cmt__AttributeCategory__c" 
    if id.startswith("a1B"):
        return "vlocity_cmt__AttributeCategory__c" 
    if id.startswith("a1W"):
        return "vlocity_cmt__Picklist__c" 
    if id.startswith("a4W"):
        return "vlocity_cmt__Picklist__c" 
    if id.startswith("a2m"):
        return "vlocity_cmt__PriceList__c" 
    if id.startswith("a4a"):
        return "vlocity_cmt__PriceList__c" 
    if id.startswith("a2j"):
        return "vlocity_cmt__PicklistValue__c"      
    if id.startswith("a3I"):
        return "vlocity_cmt__ContextDimension__c" 
    if id.startswith("a3b"):
        return "vlocity_cmt__ContextMapping__c" 
    if id.startswith("a3r"):
        return "vlocity_cmt__ContextScope__c" 
    if id.startswith("a3q"):
        return "vlocity_cmt__ContextAction__c" 
    if id.startswith("a0w"):
        return "vlocity_cmt__EntityFilter__c"   
    if id.startswith("a1o"):
        return "vlocity_cmt__Rule__c"        
    if id.startswith("a1l"):
        return "vlocity_cmt__RuleFilter__c"        
    if id.startswith("a41"):
        return "vlocity_cmt__RuleAssignment__c"        
    if id.startswith("a4x"):
        return "vlocity_cmt__OfferMigrationPlan__c"         
    if id.startswith("a4"):
        return "vlocity_cmt__OfferMigrationComponentMapping__c"       
    if id.startswith("005"):
        return "User"   
    if id.startswith("a2H"):
        return "vlocity_cmt__Catalog__c"   
    if id.startswith("a1j"):
        return "vlocity_cmt__Catalog__c"  
    if id.startswith("a04"):
        return "vlocity_cmt__CatalogProductRelationship__c"   
    if id.startswith("a5b"):
        return "vlocity_cmt__ServicePoint__c"        

    print(f" getObjectType--> {id} no related to object")
    return None
#---------------------------------

def get(id,sobjectName=None):
    if sobjectName == None:
        sobjectName = getSObjectType(id)
    # return select_wherexx_field_value_n(objectName,'Id',id)
    return query1.query(f" select fields(all) from {sobjectName} where Id='{id}' limit 200")

#---------------------------------

def create(sobjectname,object):
    call =  client.callAPI(f'/services/data/{apiVersion}/sobjects/{sobjectname}/', method="post", data=object) 
    if client.lastThreadCall()['status_code'] == 201:
        logging.info(f"Object {sobjectname} created sucesfully. Id:{call['id']}")
    else:
        logging.warning(f"Object {sobjectname} creation returned status code {client.lastThreadCall()['status_code']}")
        logging.warning(f"Response {call}")

        raise ValueError(simplejson.dumps(call, indent=4))


    return call

def checkError():
    lc = client.lastThreadCall()
    if lc['status_code'] >= 400:
        logging.error(f"Error in call.  statusCode->{client.lastThreadCall()['status_code']}")
        message = {
            'error':lc['error'],
            'errorCode':lc['errorCode']
        }
        if 'errorOther' in lc:
            message['errorOther'] = lc['errorOther']
        msg = simplejson.dumps(message, indent=4)
        logging.error(msg)
        raise ValueError(msg)

def update(id,data,sobjectname=None,getObject=True):
    if sobjectname == None:
        sobjectname = getSObjectType(id)
    
    call = client.callAPI(f'/services/data/v51.0/sobjects/{sobjectname}/{id}/',method='patch',data=data)
    checkError()
    if getObject == True:
        call = get(id)
    return call

def delete(sobjectname,id):
    if id==None:
        return None
    if type(id) is list:
        idList = ",".join(id)
        logging.info(f"deleting {sobjectname} with Id {idList}")
        return client.callAPI(f'/services/data/{apiVersion}/composite/sobjects?ids={idList}', method="delete")
    id = Id(id)
   
    logging.info(f"deleting {sobjectname} with Id {id}")
    return client.callAPI(f'/services/data/{apiVersion}/sobjects/{sobjectname}/{id}', method="delete")

def listObjects():
    return client.callAPI(f'/services/data/{apiVersion}/sobjects/')['sobjects']

def describe(sobjectName):
    return client.callAPI(f'/services/data/{apiVersion}/sobjects/{sobjectName}/describe')

def listVersions():
    return client.callAPI(f'/services/data')

def getOwner(Id):
    recordType = getSObjectType(Id)
    if recordType == None:
        return

    select = f"select OwnerId from {recordType} where Id = '{Id}' limit 50"    
    call = query1.query(select)
    OwnerId = call['records'][0]['OwnerId']

    select = f"select FIELDS(ALL) from user where Id = '{OwnerId}' limit 50"    
    user = query1.query(select)['records'][0]

    print(user['Name'])

    return call

def getNumOwnedRecords(sobject,ownerId):
    try:
        select = f"select COUNT() from {sobject} where OwnerId = '{ownerId}' "    
        call = query1.query(select)

        totalSize = call['totalSize']

        return totalSize

    except Exception as e:
        if 'INVALID_FIELD' in f"{e}":
            return -1
    
    return -2

def getNumOwnedRecords_4_ObjectList(sobjectNameList,ownerId):
    result=[]
    for obj in sobjectNameList:
        res = {
            "object":obj,
            "ownedRecords":getNumOwnedRecords(obj,ownerId)
        }
        result.append(res)

    return result

def selectOwnedRecords(sobjName,ownerId):
    select = f"select Fields(ALL) from {sobjName} where ownerId='{ownerId}'limit 50"
    call = query1.query(select)

    return call



import logging
import threading,time
import math
import sys


_counter = 0
#---------------------------------
def getCounter():
  global _counter
  return _counter
#---------------------------------
def count():
  return _counter
  
def printCounter():
  global _counter

  print(_counter)
#---------------------------------
def _testMultiple(func,iterations,processNumber,**funcArg):
  global _counter
  for a in range(iterations):
    func(processNumber=processNumber+a)
    with threadLock:
      _counter += 1
#---------------------------------
threadLock = threading.Lock()
def run(func,iterations,processStartNumber=0,threads=50,onFinished=None):
  global _counter
  _counter = 0

  try:
    for _itera in range(threads):
      processNumber = processStartNumber+ (_itera*threads)
      threading.Thread(target=_testMultiple,args=(func,iterations,processNumber), daemon=True).start()

  except:
    print ("Error: unable to start thread")
  while _counter < iterations * threads :
    logging.info(f'Iteration {_counter}  from {iterations * threads}')
    time.sleep(30)

  if onFinished!=None:
    onFinished()
    
  print("Done")

#---------------------------------
_processed = 0
_totalToProcess = 0

def processList(func,theList,threads):
  global _processed,_totalToProcess

  _processed = 0
  _totalToProcess = len(theList)
  itemsPerTh = math.ceil(_totalToProcess/threads)


  total = 0
  while total < _totalToProcess:
    end = total+itemsPerTh
    if end>_totalToProcess:
      end = _totalToProcess
    data = {
      "list" : theList[total:end]
    }

    threading.Thread(target=_processList,args=(func,theList[total:end],)).start()
    total = total + itemsPerTh

  print()

  while (_totalToProcess>_processed):
   # print(_processed)
 #   sys.stdout.write("\r%d%%" % int(100*_processed/_totalToProcess))
 #   sys.stdout.flush()
    time.sleep(1)

def progress():
    global _processed,_totalToProcess

    sys.stdout.write("\r%d%%" % int(100*_processed/_totalToProcess))
    sys.stdout.flush() 
    if (_processed==_totalToProcess):
      print()


def _processList(func,ls):
  global _processed
  for l in ls:
    func(l)
   # print('one done')
    with threadLock:
      _processed += 1
      progress()
 # print('X done')





from . import restClient, objectUtil,digitalCommerce,utils,thread
import time
import sys

def _getItemAtPath(obj,path,field='ProductCode'):
    for p in path.split(':'):
        obj = objectUtil.getSibling(obj,field,p)  
    return obj

def updateOfferField(offerDetails,path,field,value,pathField='ProductCode'):
    obj = _getItemAtPath(offerDetails,path,pathField)
    if obj == '':
        raise ValueError(f'Object does not contain element by path {path}')
    if field in obj:
        obj[field] = value
    else:
        raise ValueError(f'Object does not contain field {field} by path {path}')

    return offerDetails

def updateOfferAttribute(offerDetails,path,AttributeCategory,productAttribute,value):
    obj = _getItemAtPath(offerDetails,path)
    obj = _getItemAtPath(obj,AttributeCategory,field='Code__c')
    obj = _getItemAtPath(obj,productAttribute,field='code')

    obj['userValues'] = value
 
    return offerDetails

def getBasketProductAttributes(basket,path):
    obj = _getItemAtPath(basket,path)
    obj = obj['attributeCategories']

    return obj

def getAction(basket,path,method='addChildBasketAction'):

    lpath = path.split(':')
    xpath = ":".join(lpath[:-1])
    offer = lpath[-1]


    obj = _getItemAtPath(basket,xpath)
    action = _getItemAtPath(obj,method,field='method')

    action['params']['offer'] = offer
    action['link'] = f"/services/apexrest/{restClient.getNamespace()}{action['link']}"



    return action


def checkOffers(path=None,quantity=None):

    catalogs = digitalCommerce.getCatalogs()
    print()
    print('Catalogs in the Org:')
    utils.printFormated(catalogs,"Name:vlocity_cmt__CatalogCode__c:vlocity_cmt__IsActive__c%Active:vlocity_cmt__Description__c")
    print()

    catOffers = []
    numOffersList = []

    print("Getting offfers per catalog.")
    getOfferTimes = []
    def getOffersPerCatalog(catalog):
        if catalog['vlocity_cmt__IsActive__c'] == False:
            return 
        try:
            times = {
                'name':catalog['vlocity_cmt__CatalogCode__c'],
                'Error':'',
                "__color__":utils.CEND
            }
            offers = digitalCommerce.getOfferByCatalogue(catalog['vlocity_cmt__CatalogCode__c'])
            times['elapsed'] = restClient.getLastCallTime()
            l = len(offers) if offers is not None else 0
            numOffersList.append(l)
            times['# Offers'] = l

        except Exception as e:
            times['elapsed'] = restClient.getLastCallTime()
            times['Error'] = e.args[0]['error']
            times['__color__'] = utils.CRED
            getOfferTimes.append(times)
            return   

        if offers == None:
            times['Error'] = "Has no offers."
            times['__color__'] = utils.CYELLOW
            getOfferTimes.append(times)
            return 

        catOffer = {
            'catCode':catalog['vlocity_cmt__CatalogCode__c'],
            'offers':offers
        }
        catOffers.append(catOffer)
        getOfferTimes.append(times)

    thread.processList(getOffersPerCatalog,catalogs,1)
    utils.printFormated(getOfferTimes)
    totalOffers = sum(numOffersList)
    print()

    print(f"Getting offerDetails, createBasket, createBasket with config per offer. total offers {totalOffers}")

    offersList = []
    for catOffer in catOffers:
        for offerCount,offer in enumerate(catOffer['offers']):
            _offer = {
                'catalogCode':catOffer['catCode'],
                'offerCode':digitalCommerce.getOfferCode(offer)
            }
            offersList.append(_offer)

    thread.processList(dba,offersList,50)
    print()
    newlist = sorted(_theTimes, key=lambda d: f"{d['catalog']}{d['offerCode']}")
    utils.printFormated(newlist)

    print()

_theTimes = []
def dba(offerCodes):
    try:
        offerCode = offerCodes['offerCode']
        catalogCode = offerCodes['catalogCode']
        times = {
            'catalog':catalogCode,
            'offerCode':offerCode,
            '__color__':utils.CEND
        }

        if offerCode == 'PROMO_NOS_OFFER_007':
            print()
        details = digitalCommerce.getOfferDetails(catalogCode,offerCode)
        times['details'] = restClient.getLastCallTime()

        basket = digitalCommerce.createBasket(catalogCode,offerCode)
        times['createBasket'] = restClient.getLastCallTime()
        times['basket contextKey']=basket['cartContextKey']

        offerDetails = updateOfferField(details,'0001','Quantity',4,'lineNumber')

        basket2 = digitalCommerce.createBasketAfterConfig(catalogCode,offerDetails)
        times['afterConfig'] = restClient.getLastCallTime()

    except Exception as e:
        times['error']=e.args[0]['error']
        times['__color__'] = utils.CRED

    _theTimes.append(times)





#python3 -m unittest

import unittest
from InCli import InCli
#from InCli.SFAPI import restClient

class Test_Main(unittest.TestCase):

    def test_q(self):
        InCli.main(["-u","uormaechea.devnoscat2@nos.pt","-q", "select fields(all) from Order limit 1"])

    def test_o(self):
        InCli.main(["-u","uormaechea.devnoscat2@nos.pt","-o"])

    def test_o_name(self):
        InCli.main(["-u","uormaechea.devnoscat2@nos.pt","-o","-name","order"])

    def test_o_like(self):
        InCli.main(["-u","uormaechea.devnoscat2@nos.pt","-o","-like","XOM"])

    def test_h(self):
        InCli.main(["-h"])

    def test_check_Catalog(self):
        InCli.main(["-u","NOSDEV","-checkCatalogs"])

    def test_logs(self):
        InCli.main(["-u","NOSDEV","-logs"])

    def test_log_ls(self):
        InCli.main(["-u","NOSDEV","-logs:ls"])

    def test_log_ID(self):
        InCli.main(["-u","NOSDEV","-logs","07L3O00000DajQIUAZ"])

    def test_log_last(self):
        InCli.main(["-u","NOSDEV","-logs","-last","10"])

    def test_default(self):
        InCli.main(["-default:set","u"])
        InCli.main(["-default:set","u","NOSDEV"])
        InCli.main(["-default:get","u"])        
        InCli.main(["-logs","-last","1"])
        InCli.main(["-default:del","u"])

    def test_d(self):
        InCli.main(["-u","NOSDEV","-d"])
        InCli.main(["-u","NOSDEV","-d","Order"])
        InCli.main(["-u","NOSDEV","-d","Order:Status"])

    def test_l(self):
        InCli.main(["-u","NOSDEV","-l"])
        InCli.main(["-u","DEVNOSCAT2","-l"])


import unittest
from InCli.SFAPI import restClient,query


class Test_Query(unittest.TestCase):
    def test_query1(self):
        restClient.init('DEVNOSCAT2')
        res = query.query("select fields(all) from order limit 1")

        self.assertTrue(res['totalSize']>=0)

        print()



import os,sys,sys,gc
from xmlrpc.client import Boolean
from . import Sobjects, jsonFile,file, objectUtil,utils
import requests,threading
import logging,datetime,enum
import simplejson as json

_initializedConnections = []
_allThreadsContext = {}
_currentConnectionName = None
def _threadContext(setConnectionName=None):
    """
    Threads can share a connection. However the call stack belongs to each thread.
    When a thread calls this function, it can set the connection for the thread (if especified) or get the connection previously set if setConnectionName=None
    If a connection has not been set fot the thread, the current one (the latest one set for any previous thread) is assigned to the thread   
    """
    global _allThreadsContext,_currentConnectionName
    id = threading.get_native_id()
    if id not in _allThreadsContext:
        _currentConnectionName = setConnectionName if setConnectionName != None else _currentConnectionName
        if _currentConnectionName == None:
            utils.raiseException('ConnectionError',"No default connection for thread")
        th = {
            'connectionName': _currentConnectionName,
            'calls':[]
        }
        _allThreadsContext[id] = th
        return th
    if setConnectionName != None:   #change the connection for the thread
        _allThreadsContext[id]['connectionName'] = setConnectionName
        _currentConnectionName = setConnectionName
    return _allThreadsContext[id]

##################################################################
def _pushThreadCall(call):
    _calls = _threadContext()['calls']
    if len(_calls) == 5:
        _calls.pop(0)
    _calls.append(call)

def _updateThreadCall(call):
    _calls = _threadContext()['calls']
    _calls[-1] = call

def lastCall(field=None):
    return lastThreadCall(field)
def lastThreadCall(field=None):
    """
    Returns the last rest call data (request, response, others).
    """
    _calls = _threadContext()['calls']
    if field == None:
        return _calls[-1]
    return _calls[-1][field]

def checklastThreadCallError(caller):
    """
    Raises and Exception if the last call has an error. """
    lc = lastThreadCall()
    if 'error' in lc and lc['error'] is not None:
        utils.raiseException(lc['errorCode'],lc['error'],caller)

def getLastCallTime():
    return getlastThreadCallTime()
def getlastThreadCallTime():
    t = lastThreadCall('elapsedTime')
    return (t.microseconds + t.seconds * 1000000 )/1000

def getConfigOrgsNames():
    """
    Get all names for the org in the config file."""
    return [configOrg['name'] for configOrg in loadConfigData()['orgs']]

##################################################################

def glog():
    return logging.getLogger('root')
#/Users/uormaechea/Documents/Dev/python/Industries/input/ConnectionsParams.json
_configData = None
_configDataName = None
def setLoggingLevel(loggingLevel=logging.INFO):
    glog().level = loggingLevel

def loadConfigData():
    global _configData,_configDataName

    if _configData is not None:
        return _configData
    
    incli = os.environ.get('INCLI')
    if incli is not None:
        setConfigFile(incli)
        return _configData

    _configDataName = os.path.abspath(".incli/IncliConf.json")

    if file.exists(_configDataName) == False:
        configData = {
            "folders": {
                "input":os.path.abspath(".incli/input"),
                "debug":os.path.abspath(".incli/debug"),
                "output":os.path.abspath(".incli/output"),
                "log":os.path.abspath(".incli/logs")
            },
            "orgs": []
        }

        jsonFile.write(_configDataName,configData)

    _configData = jsonFile.read(_configDataName)

    return _configData
    #utils.raiseException('NoConfigFile',"No config file has been defined.")

def getConfigVar(name):
    cd = loadConfigData()
    if name in cd:
        return cd[name]
    return None

def setConfigVar(name,value):
    cd = loadConfigData()
    cd[name] = value
    jsonFile.write(_configDataName,cd)

def delConfigVar(name):
    cd = loadConfigData()
    try:
        del cd[name]
    except KeyError:
        glog().info(f'Variable {name} is not set.')
        return
    jsonFile.write(_configDataName,cd)
    glog().info(f'Variable {name} deleted.')

def saveOrg_inConfigFile(orgName,instance_url,token=None):
    """to save in the config file Guest or Bearer Org connection params."""
    isGuest = True if token == None else False

    cd = loadConfigData()

    for org in cd['orgs']:
        if org['name'] == orgName:
            org['instance_url'] = instance_url
            if token != None:
                org['bearer'] = token
            jsonFile.write(_configDataName,cd)
            return
    
    org =     {
        "name":orgName,
        "instance_url": instance_url,
        "nameSpace": "vlocity_cmt"
    }
    if token!=None:
        org['bearer'] = token

    cd['orgs'].append(org)
    jsonFile.write(_configDataName,cd)

def deleteOrg_inConfigFile(orgName):
    cd = loadConfigData()
    cd2 = [i for i in cd['orgs'] if not (i['name'] == orgName)]
    cd['orgs'] = cd2
    jsonFile.write(_configDataName,cd)
    loadConfigData()

def setLoggingLevel(level=logging.INFO):
    log = logging.getLogger('root')
    logging.basicConfig()
    log.setLevel(level)

def setConfigFile(configFile):
    """
    Set the config file to use, and the log level"""
    global _configData,_configDataName

    if file.exists(configFile):
        _configData = jsonFile.read(configFile)
    else:
        utils.raiseException("NO_CONFIG",f"Cannot open the configuration file <{configFile}>, please provide a valid configuration file (path and name).")

    _configDataName = configFile

sfdx_lock = threading.Lock()
####CONECTION
def init(userName_or_orgAlias,connectionName=None):
    with sfdx_lock:
        inConf = False
        if userName_or_orgAlias == None:
            userName_or_orgAlias = getConfigVar('u')
            if userName_or_orgAlias == None:
                utils.raiseException('Configuration',"No userName or Org Alias specified. Please specify a user name or org alias.")
            inConf = True    

        connectionName = connectionName if connectionName is not None else userName_or_orgAlias
        if _checkAndSetConnectionIfExists(connectionName):
            return

        try:
            success,outputs = utils.executeCommandParse(["sfdx","force:org:display","-u", userName_or_orgAlias])

        except Exception as e:
            if e.strerror == 'No such file or directory':
                utils.raiseException('SFDXError',"SFDX is not installed or it is not accesible.",other='https://developer.salesforce.com/docs/atlas.en-us.sfdx_setup.meta/sfdx_setup/sfdx_setup_install_cli.htm')
            else:
                utils.raiseException('SFDXError',e.strerror)

        if success is False:
            error = outputs.stderr.split('force:org:display')[1]
            if 'No AuthInfo found' in error:
                addText = " set in the configuration " if inConf==True else ''
                utils.raiseException('ConnectionError',f"{error}. Please authorize the org for the {userName_or_orgAlias}{addText}: sfdx auth:web:login",other="Check Connection status: sfdx force:org:list --verbose --all")
            else:
                utils.raiseException('SFDX Error',error,other='')

        obj  = {}
        for output in outputs:
            obj[output['KEY']] = output['VALUE']

        if obj['Connected Status'] != 'Connected':
            utils.raiseException("ConnectionStatus",f"Connected Status for client Id {userName_or_orgAlias} is {obj['Connected Status']}",other=f"Execute the following command to refresh the token.  -")

        assert(connectionName!=None)
        assert(obj['Instance Url']!=None)
        assert(obj['Access Token']!=None)

        _initMain(connectionName,obj['Instance Url'],obj['Access Token'])

def initWithToken(name,url,token=None,input=None,output=None,debug=None):
    _initMain(name,url,token,input,output,debug)

def initWithConfig(orgName,isGuest=False,connectionName=None)->Boolean:
    """
    Reads the ConnectionsParams.json configuration specified by environment. If isGuest=False it will authenticate with Salesforce and obtain the token and url. If isGuest=True, it will not authenticate and the url must be provided in the ConnectionsParams.

    - environment: a string identifying the connection in the ConnectionsParams file.
    - isGuest: if True, authentication will not be performed and the ConnectionsParam requires to provide the url. 
    - name: the name of the connection. If not provided name=environment. Used when 2 connections are established for the same environment. 
    - configFolder: The folder with the ConnectionsParams file.
    - configFileName: the name of the config file. 
    - outputFolder: folder for the debuglogs
    - outputFolder: output folder. 
    """

    if orgName not in getConfigOrgsNames():
        utils.raiseException("NO_ORG",f"Org name {orgName} is not valid. Does not exist in the Configuration file.")

    if connectionName == None:
        connectionName = orgName

    orgConfig = objectUtil.getSibling(_configData['orgs'],"name",orgName).copy()

    url = orgConfig['instance_url'] if 'instance_url' in orgConfig else None
    token = orgConfig['bearer'] if 'bearer' in orgConfig else None

    if token is None:
        if 'login' in orgConfig:
           # raise ValueError(f"Environment connection parameters missing login parameters. {connectionName}")      
            url,token = _authenticate(orgConfig['login'],orgConfig['isSandBox'])

    if token is None and url is None:
        raise ValueError(f"Provide a instance_url for guest users (onboarding). {connectionName}") 

    _initMain(connectionName,url=url,token=token)

    return True

def _checkAndSetConnectionIfExists(connectionName):
    for con in _initializedConnections:
        if con['connectionName'] == connectionName:
            _threadContext(connectionName)
            glog().info(f"Connection {connectionName} set.")
            return True
    return False

def _initMain(name,url,token=None,input=None,output=None,debug=None):
    if _checkAndSetConnectionIfExists(name):
        return

    loadConfigData()

  #  currentDir = os.getcwd()
    connection = {
        'connectionName':name,
        'isGuest': True if token is None else False,
        'access_token':token,
        'instance_url':url,
        'input':input if input is not None else _configData['folders']['input'],
        'output':output if input is not None else _configData['folders']['output'],
        'debug':debug if input is not None else _configData['folders']['debug'],
        'log':debug if input is not None else _configData['folders']['log'],

        'nameSpace':'vlocity_cmt'
    }
    _initializedConnections.append(connection)
    _threadContext(connection['connectionName'])

    glog().info(f"Connection {connection['connectionName']} initialized.")

def getCurrentThreadConnection():
    """
    Retrieves the connectionParams for the current org. 
    """
    global _initializedConnections
    connectionName = _threadContext()['connectionName']
    if connectionName == None:
        utils.raiseException('ConnectionError',"No connection has been established for current thread.",other="Make sure the connection is established.")
    connection = [con for con in _initializedConnections if con['connectionName']==connectionName][0]
    #connection = objectUtil.getSibling(_initializedConnections,"name",name)
    return connection

def getNamespace():
    connection = getCurrentThreadConnection()
    return connection['nameSpace']
def inputFolder():
    return _gerFolder('input')
def outputFolder():
    return _gerFolder('output')
def debugFolder():
    return _gerFolder('debug')
def logFolder():
    return _gerFolder('log')
def _gerFolder(name):
    folder =  getCurrentThreadConnection()[name]
    if folder[-1] != '/':
        folder = folder + '/'
    if file.exists(folder) == False:
        os.makedirs(folder)
    return folder

#################################################################################################################################
def _authenticate(login,isSandbox):
    if 'isSandBox' == None:
        raise ValueError(f"Environment connection parameters missing isSandBox field.")

    headers = {
        'Content-type': 'application/x-www-form-urlencoded'
    }
    server = 'test' if isSandbox else 'login'
    _threadContext('oauth2')
    call = requestRaw(url=f"https://{server}.salesforce.com/services/oauth2/token",method='post', parameters= login,headers=headers)
    
    lc = lastThreadCall()
    if lc['error'] is not None:
        utils.raiseException(lc['errorCode'],lc['error'],other=lc['errorOther'])

   # connection['access_token'] = call["access_token"]
   # connection['instance_url'] = call["instance_url"]

    glog().info('getting token')
    glog().info(f"Authenticated. Instance URL is {call['instance_url'] }")
    glog().info(f"Authenticated. Bearer token {call['access_token']}")

    return call["instance_url"],call["access_token"]

def requestRaw(url,action=None,method = 'get',parameters = {},data={},headers={},access_token=None):
    """
    Basic request Call. 
    
    No need to perform init(), as it does not use the connectionParams and all information needs to be provided. 
    Method parameters are self explanatory. 

    """
    
    completeUrl = url
    if action!=None:
        completeUrl = url + action

    allheaders = {
        'Content-type': 'application/json',
        'User-Agent':'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.72 Safari/537.36',
        'Accept-Encoding': 'gzip, deflate, br'
    }

    if headers:
        allheaders.update(headers)

    if access_token != None:
        allheaders['Authorization'] = 'Bearer %s' % access_token

    method = method.lower()

    call = {
        'action':action,
        'url':url,
        'parameters':parameters,
        'method':method,
        'data':data,
        'error':None,
        'errorCode':None,
        'callTime':datetime.datetime.now()
    }

    glog().debug(f"Current Connection {_currentConnectionName}")
    _pushThreadCall(call)

    r = None

    glog().debug(f"{call['url']}{call['action']}")
    try:
        if method in ['get','delete']:
            r = requests.request(method, completeUrl, headers=allheaders, params=parameters, timeout=300)
            r.raise_for_status()
        elif method in ['post', 'patch','put']:
            r = requests.request(method, completeUrl, headers=allheaders, json=data, params=parameters, timeout=300)
            r.raise_for_status()
        else:
            utils.raiseException('NO_METHOD',f'Method {method} is not implemented in restClient.','restClient')

    except requests.exceptions.HTTPError as errh:
        call['error'] = f'serverResponse: {r.text}'
        call['errorCode'] = f"HTTPs Error: {r.status_code}"
        call['errorOther'] = f"httpError:':{errh}"
    except requests.exceptions.ConnectionError as errc:
        call['error'] = {'ConnectionError':f"{errc}"}
        call['errorCode'] = f"ConnectionError"
    except requests.exceptions.Timeout as errt:
        call['error'] = {'Timeout':f"{errt}"}
        call['errorCode'] = f"Timeout Error"    
    except requests.exceptions.RequestException as err:
        call['error'] = {'RequestException':f"{err}"}
        call['errorCode'] = f"RequestException"

    if r != None:
        glog().debug(f'Debug: API {method} call: {r.url}  status Code:{r.status_code}' ) 
        
        call['responseTime'] = datetime.datetime.now()
        call['status_code'] = r.status_code
        call['elapsedTime'] = r.elapsed
        call['elapsedTimeCall'] = r.elapsed
        call['deltaTime'] = call['responseTime'] - call['callTime']

        if r.status_code < 300 :
            if r.text == '':
                call['response'] = ''
            else:
                try:
                    call['response'] = r.json()
            
                except Exception as e:
                    glog().debug(f"warn. Response is not json  --> {e}")
                    call['response'] = {}
                    call['response'] = r.text
                
        else:
            glog().warn('API error when calling %s : %s' % (r.url, r.content))
            call['response'] = call['error']

    _updateThreadCall(call)

    return call['response']

def requestWithConnection(action,  parameters = {}, method = 'get', data = {},headers = {}):
    """
    Performs a request using the current connection as configured in the file. 
    """
    connection = getCurrentThreadConnection()

    if connection == None:
        raise ValueError('restClient current org is not set. Have you init restClient?')

    return requestRaw(  url=connection['instance_url'],
                        action=action, 
                        parameters=parameters , 
                        method = method , 
                        data = data ,
                        headers  =headers,
                        access_token=connection['access_token'] if 'access_token' in connection else None
                        )

def callAPI_multi(action,params={} , method = 'get', data = {},headers={}):
    done = False

   # parameters = params
    totalElepsedTime = datetime.timedelta(microseconds=0)
    totalCalls = 0

    while done==False:
        call = requestWithConnection(action,parameters = params,method=method, data=data, headers=headers)

        totalElepsedTime = lastThreadCall('elapsedTime') + totalElepsedTime
        totalCalls = totalCalls + 1

        if  call == None or lastThreadCall('status_code')>=300:
            break
        
        glog().debug(f"callAPI_multi: <{action}>  ts:{lastThreadCall('elapsedTime')}") 

        #For chainable integration procedures
        if 'IPResult' in call and 'vlcStatus' in call['IPResult'] and call['IPResult']['vlcStatus'] == 'InProgress':
            data['input'] = "{}"
            data['options'] = json.dumps({
                'vlcMessage':None,
                'vlcIPData':call['IPResult']['vlcIPData'],
                'vlcStatus':'InProgress'
            })


        #For digital commerce
        elif 'nexttransaction' in call and type(call)==dict:
            multiTransactionKey = call['nexttransaction']['rest']['params']['multiTransactionKey']
            data['multiTransactionKey'] = multiTransactionKey  
        else:
            done = True

    lc = lastThreadCall()
    lc['elapsedTime'] = totalElepsedTime
    lc['totalCalls'] = totalCalls    
    _updateThreadCall(lc)

    return call

#def callAPI(action, parameters = {}, method = 'get', data = {}, headers={}):    
#    call = callAPI_multi(action,parameters,method,data,headers)
#    return call
##-------------------------------------------------------------------------------------------
# stores the request input and the reply into files in the output directory
# The file name is provided by the calling function tree "debugFile(filename)", or calculated if not provided
# the request add _req to the file name, while the reply adds _rep to the file name
# The reply can be processed -> change the json to take out data, or compute additional fields before storing it
#def callAPI_debug(action, parameters = {}, method = 'get', data = {}, headers={}):    
    return callAPI(action,parameters,method,data,headers)

def callAPI(action, parameters = {}, method = 'get', data = {}, headers={}):    
    call = callAPI_multi(action,parameters,method,data,headers)
    return call

def callSave(logFileName,logRequest=False,logReply=True,timeStamp=False,responseProcessing=None,requestProcessing=None):
    if logRequest == False and logReply == False:
        return  

    now =f"{utils.datetimeString_now('%H:%M:%S')}--"
    if  timeStamp == False:
        now = ''
    filename = f"{now}{logFileName}"

    lc = lastThreadCall()

    if logRequest == True:
        fn = f'{filename}_req'
        filepath = writeFile(fn,lc['data'])
        lc['requestFilePath'] = filepath

    if logReply == True:
        if requestProcessing != None:
            requestProcessing(lc['data'])
        
        fn = f'{filename}_res'
        try:
            filepath = writeFile(fn,lc['response'])
        except Exception as e:
            filepath = writeFile(fn,str(lc['response']))  
        lc['responseFilePath'] = filepath

    _updateThreadCall(lc)

def writeFile(filename,content):
    connection = getCurrentThreadConnection()

    try:
        cont = content
        if (type(content) is dict or type(content) is list ):
            return jsonFile.writeFileinFolder(outputFolder(),filename,cont)
        else:
            cont = json.dumps(cont)

    except ValueError as e:
        return file.writeFileinFolder(outputFolder(),filename,str(content))        
        
def initTest():
    init("DEVNOSCAT2",configFolder="../input")     




from . import restClient,query,digitalCommerceUtil,file,utils,objectUtil
import simplejson,logging

def queryLogRecords(logUserId=None,limit=50):
    whereUSer = f" where logUserId='{logUserId}' " if logUserId is not None else ''

    call = query.query(f"Select Id,LogUserId,LogLength,LastModifiedDate,Request,Operation,Application,Status,DurationMilliseconds,StartTime,Location,RequestIdentifier FROM ApexLog  {whereUSer} order by LastModifiedDate desc limit {limit}")
    return call

def queryLogData(logId):
#    log = query.queryRecords(f"Select fields(all) FROM ApexLog where Id ='{logId}' limit 1")
    log = query.queryRecords(f"Select Id,LogUserId,LogLength,LastModifiedDate,Request,Operation,Application,Status,DurationMilliseconds,StartTime,Location,RequestIdentifier FROM ApexLog where Id ='{logId}' limit 1")

    if log == None or len(log)==0:
        utils.raiseException(errorCode='NO_LOG',error=f'The requested log <{logId}> cannot be found in the Server.',other=f"No record in ApexLogwith Id {logId}")    
    log = log[0]
    logLine = f"""LOGDATA:    Id: {log['Id']}   LogUserId: {log['LogUserId']}    Request: {log['Request']}  Operation: {utils.CGREEN}{log['Operation']}{utils.CEND}    lenght: {log['LogLength']}    duration: {log['DurationMilliseconds']}    startTime: {log['StartTime']} 
LOGDATA:    app: {log['Application']}      status: {log['Status']}     location: {log['Location']}     requestIdentifier: {log['RequestIdentifier']}
    """    

    action = f"/services/data/v56.0/sobjects/ApexLog/{logId}/Body/"
    callbody = restClient.callAPI(action)
    callbody = logLine + callbody
    return log,callbody

def parseLog(logId=None,printLimits=True,userDebug=True,lastN=1,logUserId=None):
    body = None

    if logId is not None:
        filename = f"{restClient.logFolder()}{logId}.log"
        if file.exists(filename) == True:
            body = file.read(filename)
       #     body = None

    if body is None:
        if logId is None:
            whereUSer = f" where logUserId='{logUserId}' " if logUserId is not None else ''
            logIds = query.queryFieldList(f"Select Id FROM ApexLog {whereUSer} order by LastModifiedDate desc limit {lastN}")
            if logIds == None or len(logIds)==0:
                utils.raiseException(errorCode='NO_LOG',error=f'No logs can be found. ')
            for logId in logIds:
                print()
                print(f"Parsing log Id {logId}")
                parseLog(logId)
            #    input("Click for next")
            return  
        else:
            logRecord,body = queryLogData(logId)

    if body == None or len(body)==0:
        utils.raiseException(errorCode='NO_LOG',error=f'The requested log <{logId}> cannot be found. ')
    filename = f"{restClient.logFolder()}{logId}.log"

    file.write(filename,body)
    parse(body,printLimits=printLimits,userDebug=userDebug)

def printLogs(logUserId=None,limit=50):
    logs = queryLogRecords(logUserId,limit=limit)
    logs = utils.deleteNulls(logs,delSystemFields=False)
    utils.printFormated(logs)

def delta(obj,field):
    return obj[field][1] - obj[field][0] if len(obj[field]) > 1 else 0

def setTimes(line,obj=None,field=None,value=None,chunkNum=None,type=None):
    def addList(obj,field,val):
        if field in obj:
            obj[field].append(val)
        else:
            obj[field] = [val]

    chunks = line.split('|')

    if obj == None:
        obj = {
            'type' : type,
            'ident' : _context['ident'],
            'exception' :False
        }
       
        if len(chunks)>3:
            obj['Id'] = chunks[3]

    addList(obj,'lines',line)
    addList(obj,'CPUTime',_context['DEF:CPU time'])
    addList(obj,'SOQLQueries',_context['DEF:SOQL queries'])
    addList(obj,'cmtCPUTime',_context['CMT:CPU time'])
    addList(obj,'cmtSOQLQueries',_context['CMT:SOQL queries'])
    addList(obj,'totalQueries',_context['totalQueries'])
    addList(obj,'time',chunks[0].split(' ')[0])
    addList(obj,'timeStamp',int ((chunks[0].split('(')[1]).split(')')[0]))

    if _context['exception'] == True and obj['exception'] == False:
        obj['exception'] = True
        _context['exception'] = False

    if obj['type'] is None:
        print()

    if field is not None:
        obj[field] = chunks[chunkNum] if value is None else value

    if _context['timeZero'] == 0:
        _context['timeZero'] = obj['timeStamp'][0]

    obj['elapsedTime'] = obj[f'timeStamp'][0] - _context['timeZero']

    return obj

_context = {
    'totalQueries' : 0,
    'timeZero':0,
    'ident':0,
    'DEF:SOQL queries' : 0,
    'DEF:CPU time' : 0,
    'CMT:SOQL queries' : 0,
    'CMT:CPU time' : 0,
    "exception":False
}

#######################################################################
def resetContext():
    _context['totalQueries'] = 0
    _context['timeZero'] = 0
    _context['ident'] = 0
    _context['DEF:SOQL queries'] = 0
    _context['DEF:CPU time']=0
    _context['CMT:SOQL queries'] = 0
    _context['CMT:CPU time']=0
    _context['exception'] = False
    _context['previousElapsedTime'] = 0
    _context['previousCPUTime'] = 0
    _context['previousIsLimit'] = False
    _context['prevTimes'] = {
        "0":[0,0]
    }
    _context['firstLineIn'] = True
    _context['firstLineOut'] = True

    

def parse(log,printLimits=True,userDebug=True):
   # log = file.read(filepath=filepath)
    _context['printLimits'] = printLimits
    _context['userDebug'] = userDebug

    lines = log.splitlines()

    ident = 0

    SOQL = {}

    #totalQueries = 0

    methodsList = []
    constructorList = []
    codeUnitsList = []
    constructorList = []
    dmlList=[]
    limits = None

    exceptionDict = None

    limitsDict = {}

    #workflow = {}

    debugList = []
    multiLine = None #for multiline commands

    resetContext()

    print()

    for num,line in enumerate(lines):
        chunks = line.split('|')

        if _context['firstLineIn'] == True:
            if 'LOGDATA:' in line:
                print(line)
                continue
            else:
                _context['firstLineIn'] = False
                levels = line.strip().split(' ')[1].replace(',','=').replace(';','  ')
                print(levels)
                print()
                continue

        if len(chunks)>1 and chunks[1] in ['HEAP_ALLOCATE','STATEMENT_EXECUTE','USER_INFO','VARIABLE_SCOPE_BEGIN']:
            continue

        if '|SYSTEM_MODE_EXIT|' in line:
            nop=1

        if '|USER_INFO|' in line:
            setTimes(line)

        if '|' in line:   #This is a new line always. 
            multiLine = None
        
        if exceptionDict is not None:
            if multiLine != 'EXCEPTION':   #This is a new line without the exeption
                debugList.append(exceptionDict.copy())
                exceptionDict = None
                multiLine = None
            else:
                if line != '':
                    exceptionDict['line'] = exceptionDict['line'] + line
        if '|EXCEPTION_THROWN|' in line or 'FATAL_ERROR' in line:
            exceptionDict = {
                'line': line,
                'type': 'EXCEPTION'
            }
            _context['exception'] = True
            multiLine = 'EXCEPTION'

        if '|VARIABLE_ASSIGNMENT|' in line:
            if len(chunks) >= 5:
                if 'ExecutionException' in chunks[4]:
                    obj = setTimes(line,type='VAR_ASSIGN')
                    obj['type'] = 'VAR_ASSIGN'
                    obj['subType'] = 'EXCEPTION'
                    obj['string'] = chunks[4]
                    obj['apexline'] = chunks[2][1:-1] if chunks[2]!='[EXTERNAL]' else 'EX'

                    debugList.append(obj)
 
        if '|LIMIT_USAGE|' in line:
            if '|SOQL|' in line:
                _context[f'DEF:SOQL queries'] = chunks[4]

        if '|LIMIT_USAGE_FOR_NS|' in line:
            limits = chunks[2]
            if limits == '(default)':
                limits = 'DEF:'
            elif limits == 'vlocity_cmt':
                limits = 'CMT:'
            else:
                limits = f"{limits}:"

            limitsDict = setTimes(line,type='LIMITS')
            limitsDict['limitType'] = limits

        if limits is not None and line == ' ':
            limits = None
        
        if limits is not None:
            chunks = line.split(' ')
            if 'SOQL queries' in line:
                _context[f'{limits}SOQL queries'] = chunks[6]
            if 'CPU time' in line:
                _context[f'{limits}CPU time'] = chunks[5]

        if limits is not None and line =='':
            setTimes(limitsDict['lines'][0],limitsDict)
            s = f"limits-{limits.lower()} {limitsDict['time'][0]}"
            if _context['printLimits'] == True:
                debugList.append(limitsDict)
            limits = None   


        if '*** getCpuTime() ***' in line:
            chs = chunks[4].split(' ')
            _context[f'DEF:CPU time'] = chs[4]
        if '*** getQueries() ***' in line:
            chs = chunks[4].split(' ')
            _context[f'DEF:SOQL queries'] = chs[4]

        if '|USER_DEBUG|' in line:
            multiLine = 'DEBUG'
        if multiLine == 'DEBUG':
            if '|' in line:
                obj = setTimes(line,type='DEBUG')
                obj['type'] = 'DEBUG'
                obj['subType'] = chunks[3]
                obj['string'] = chunks[4]
                obj['apexline'] = chunks[2][1:-1]

            else:
                obj = debugList[-1].copy()
                obj['string'] = line
        #    if _context['userDebug'] == True:
            debugList.append(obj)

    
        if '|WF_RULE_EVAL' in line:
            if 'BEGIN' in chunks[1]:
                workflow = setTimes(line,field='method',value=chunks[2])

            if 'END' in chunks[1]:
                setTimes(line,workflow)
        if '|WF_ACTION|' in line:
            workflow['action'] = chunks[2]

        if '|CONSTRUCTOR_' in line:
            if 'ENTRY' in chunks[1]:      
                obj = setTimes(line,field='method',value=chunks[5],type='CONSTRUCTOR')
                constructorList.append(obj)
                increaseIdent()    
                debugList.append(obj)


            if 'EXIT' in chunks[1]:    
                decreaseIdent()
                obj = delFromList(constructorList,'method',chunks[5])
                setTimes(line,obj)   
         
        if '|CODE_UNIT_' in line:
            if 'STARTED' in chunks[1]:
                obj = setTimes(line,type='CODE_UNIT')
                obj['trigger'] = chunks[3]  if len(chunks)>4 else ''
                obj['method'] = chunks[4] if len(chunks)>4 else chunks[3]
                codeUnitsList.append(obj)
                increaseIdent()
                debugList.append(obj)

            if 'FINISHED' in chunks[1]:
                decreaseIdent()
                obj = delFromList(codeUnitsList,'method',chunks[2])
                setTimes(line,obj)

        if '|DML_' in line:
            if 'BEGIN' in chunks[1]:
                obj = setTimes(line,type="DML")
                obj['OP'] = chunks[3]
                obj['Type'] = chunks[4]
                obj['Id'] = chunks[2]
                obj['apexline'] = chunks[2][1:-1]
                dmlList.append(obj)
                increaseIdent()
                debugList.append(obj)

            if 'END' in chunks[1]:
                obj = delFromList(dmlList,'Id',chunks[2])
                setTimes(line,obj)
                decreaseIdent()

        if 'SOQL_EXECUTE_' in line:
            if 'BEGIN' in chunks[1]:
                SOQL = setTimes(line,field='query',value=chunks[4],type="SOQL")
                SOQL['object'] = chunks[4].lower().split(' from ')[1].strip().split(' ')[0]
                SOQL['apexline'] = chunks[2][1:-1]

                a= 1
                debugList.append(SOQL)

            if 'END' in chunks[1]:
                _context['totalQueries'] = _context['totalQueries'] + 1
                setTimes(line,SOQL)
                SOQL['rows'] = chunks[3].split(':')[1]
                if len(SOQL['query']) > 180:
                    if 'from' in SOQL['query'].lower():
                        SOQL['query'] = SOQL['query'].replace('from',"FROM")
                        SOQL['query'] = SOQL['query'].replace('From',"FROM")

                    SOQL['query'] = f"{SOQL['query'].split(' ')[0]} ... FROM {SOQL['query'].split('FROM')[1]}"

        if '|METHOD_' in line:
            if len(chunks)<4:
                print(line)
                continue

            operation = chunks[1]
            method = getMethod(line)

            if 'ENTRY' in operation:
                obj = setTimes(line,type='METHOD')
                obj['method'] = method
                obj['apexline'] = chunks[2][1:-1] if chunks[2]!='[EXTERNAL]' else 'EX'
                debugList.append(obj)

                if '.getInstance' in method:
                    pass
                else:
                    methodsList.append(obj)
                    increaseIdent()
            else:
                obj = delFromList(methodsList,'method',method)

                if obj is not None:
                    decreaseIdent()
                    setTimes(line,obj)

                else:
                    obj = setTimes(line,type='NO_ENTRY')

    for d in debugList:
        printParsedLog(d)
    print()


 #  utils.printFormated(logsList)
def getMethod(line):
    chunks = line.split('|')
    if len(chunks) == 4:
        method = chunks[3]
    else:
        method = chunks[4]
    if '(' in method:
        method = method.split('(')[0]
    return method

def getTime(line):
    chunks = line.split('|')
    return chunks[0].split(' ')[0]

def getTimeStamp(line):
    chunks = line.split('|')
    return int ((chunks[0].split('(')[1]).split(')')[0])

def increaseIdent():
    _context['ident'] = _context['ident'] + 1

def decreaseIdent():
    _context['ident'] = _context['ident'] - 1

def emptyString(size,char=' ',ident=None):
    str = ''
    if ident is None:
        ident = _context['ident']
    length = ident * size
    for x in range(length):
        str = str + char  
    return str     

def rootString():
    str = ''
    length = _context['ident'] 
    if length == 0:
        return ''
    for x in range(length-1):
        str = str + '⎮'
    str = str + '⌈' 

    return str     

bufVal=''
buffer=''
previousEnd = False

prevTS = 0


def printParsedLog(d):

    CEND    = '\33[0m'
    CRED    = '\33[31m'
    CWHITE  = '\33[37m'
    CBLUE   = '\33[34m'
    CGREEN  = '\33[32m'
    CYELLOW = '\33[33m'

    Cinit = CWHITE

    _plimit=' '
    if d['type'] == 'EXCEPTION':
        print(CRED + d['line'] + CEND)
        return

    if d['type'] == 'LIMITS':
        _context['previousIsLimit'] = True
        return
    if _context['previousIsLimit'] == True:
        _plimit = '*' 
        _context['previousIsLimit'] = False

    level = d['ident']

    _type = d['type']
    if _type == 'DEBUG':
        _type = f"{d['type']}-{d['subType']}"
        Cinit = CRED if d['subType'] == 'ERROR' else CGREEN

    if _type == 'VAR_ASSIGN':
        if d['subType'] == 'EXCEPTION':
            Cinit = CRED
        else:
            return

    if d['type'] == 'SOQL':
        Cinit = CBLUE
    if d['type'] == 'DML':
        Cinit =  CYELLOW

    i = f"{emptyString(1,' ',level)}."
    i= level
    identation = f"{emptyString(3,' ',level)}"

    val = d['query'] if d['type'] == 'SOQL' else ''
    val = d['method'] if d['type'] == 'METHOD' else val
    val = d['method'] if d['type'] == 'CONSTRUCTOR' else val
    val = d['method'] if d['type'] == 'CODE_UNIT' else val
    val = f"{d['OP']} {d['Type']}" if d['type'] == 'DML' else val
    val = f"{d['limitType']} " if d['type'] == 'LIMITS' else val
    val = d['string'] if d['type'] == 'DEBUG' else val
    val = d['string'] if d['type'] == 'VAR_ASSIGN' else val

    if val == '':
        print()
    
    val = Cinit +f"{identation}{val}"

    _apexline = d['apexline'] if 'apexline' in d else ''

    _totalQueriesTrace = delta(d,'totalQueries') 
    spacer = '_' if d['type'] == 'SOQL' else '.'
    _totalQueriesTrace = f"{level:2}:{emptyString(1,spacer,level)}{_totalQueriesTrace}" if _totalQueriesTrace >0 else ' '

    _cpuTime0 = int(d['CPUTime'][0])
    _cpuTime1 = int(d['CPUTime'][1]) if len(d['CPUTime']) >1 else ''
    _timeStamp1 = d['timeStamp'][1] if len(d['timeStamp'])>1 else d['timeStamp'][0]

    _totalQueries0 = d['totalQueries'][0]
    _totalQueries1 = d['totalQueries'][1] if len(d['SOQLQueries']) >1 else _totalQueries0
    _totalQueriesD = _totalQueries1-_totalQueries0

    _cpuPrevD = _cpuTime0 - int(_context['previousCPUTime'])

    _l = f"{level}"
    prev = _context['prevTimes'][_l][1] if _l in _context['prevTimes'] else _context['prevTimes'][f"{level-1}"][0]
    _elapsedPrevD = d['timeStamp'][0] - prev
    _context['prevTimes'][_l] = [d['timeStamp'][0],_timeStamp1]
    _elapsedPrevD = ms(_elapsedPrevD)

    _exp = "!" if d['exception'] == True else ''

    _sql2 = d['SOQLQueries'][1] if len(d['SOQLQueries'])>1 else d['SOQLQueries'][0]
    _sqlcmt2 = d['cmtSOQLQueries'][1] if len(d['cmtSOQLQueries'])>1 else d['cmtSOQLQueries'][0]

    _context['previousCPUTime'] = _cpuTime0
    _context['previousElapsedTime']  = d['elapsedTime']

    if _context['firstLineOut'] == True:
        print(f"{' ':15}|{'time(ms)':10}|{'elapsed':10}|{'time1(ns)':12}|{'time2(ns)':12}|{'ExecTime':10}|{'QueriesStack':15}|{'Qd':4}|{'Qt':4}|{'L':1}|{'cpuD':6}|{'CPUin':6}|{'CPUout':6}|{'Q':3}|{'Qcm':3}|{'type':12}|{'E':1}|{'al':4}{'':50}")
        _context['firstLineOut'] = False

    print(f"{i:15} {ms(d['elapsedTime']):10}|{_elapsedPrevD:8}|{d['timeStamp'][0]:12}|{_timeStamp1:12}|{ms(delta(d,'timeStamp')):10}|{_totalQueriesTrace:15}|{_totalQueriesD:4}|{_totalQueries1:4}|{_plimit:1}|{_cpuPrevD:6}|{_cpuTime0:6}|{_cpuTime1:6}|{_sql2:3}|{_sqlcmt2:3}|{_type:12}|{_exp:1}|{_apexline:>4}|{val[:150]:50}"+CEND)


logsList = []
def ms(val):
    return f"{val/1000000:10.2f}"
def printIdent(string,ident):
    str = ''
    for x in range(_context['ident'] * 3):
        str = str + ' '
    print(str + string)


def delFromList(theList,field,value):
    for i in range(len(theList)):
        obj = theList[len(theList)-i-1]

        if obj[field] == value:
            theList.pop(len(theList)-i-1)
            return obj

    for i in range(len(theList)):
        obj = theList[len(theList)-i-1]

        if value in obj[field]:
            theList.pop(len(theList)-i-1)
            return obj

    return None    




import json
from .SFAPI import Sobjects, query,restClient,jsonFile,utils,objectUtil,debugLogs,digitalCommerceUtil
import logging
import simplejson,sys
from operator import itemgetter
from inspect import getmembers, isfunction
import traceback


varsFile = 'confIcli.json'
vars = None


def readVars(field=None):
    global vars
    vars = jsonFile.read(varsFile)

    if field != None:
        return vars[field]
    return vars

def setVar(name,value):

    vars = jsonFile.read(varsFile)
    vars[name]=value
    jsonFile.write(varsFile,vars)
    print(simplejson.dumps(vars, indent=4))

def getParam(argv,key,i=1):
    if key not in argv:
        return None

    ind = argv.index(key)
    if (ind+i <= (len(argv)-1)):
        arg = argv[ind+i]
        if arg[0] == '-':
            return None
        return arg
    return None

def listEnvs():
    cons = restClient.getConfigOrgs()
    cons = sorted(cons,key=itemgetter('name'))
    utils.printFormated(cons,fieldsString='name:isSandBox:instance_url:login.username:login.password:login.bearer')
    env = readVars('environment')
    print(f"Current Environment is {env}")

def _queryAndPrint(q,fields=None,systemFields=True,nullFields=False):

    res = query.queryRecords(q)
    if fields != None:
        res = utils.deleteNulls(res,systemFields==False,nullFields==False)
        if fields=='all':
            utils.printFormated(res)
        else:
            utils.printFormated(res,fields)
    else:
        res = utils.deleteNulls(res,systemFields==False,nullFields==False)
        print(simplejson.dumps(res, indent=4))
    
    print()
    print(f"Null values -> {nullFields}  systemFields --> {systemFields}")

def option_q(args):
    if '-h' in args:
        help = """
        -q "select..." query to execute. 
            -null - will print null values as well
            -system - will print the system fields
            -fields "a:b:c:..."
            -fields all"""
        return help

    #else:
    q = getParam(args,'-q')

    fields = getParam(args,'-fields') if '-fields' in args else None    
    nullFields = True if '-null' in args else False
    systemFields = True if '-system' in args else False

    connectionInit(args)

    _queryAndPrint(q,fields=None,systemFields=True,nullFields=False)

def option_checkCatalogs(args):
    if '-h' in args:
        help = """
        -checkCatalogs 
            Checks the catalogs. Gets all catalogs, does a getOffers, getOfferDetails, basketwithoutconfig and basket with config
            path: the path to a child product to configure. Path = ... ProductCode ... ProdcutCode
            quantity: the quantity for the object""" 
        return help

    connectionInit(args)

   # path = getParam(args,'-checkCatalogs')
   # quantity = getParam(args,'-checkCatalogs',2)

    digitalCommerceUtil.checkOffers()

    print()

def option_d(args):
    help = """
        -d objectName --> Describe an object
            -d objectName:fieldName --> describe a field in the objec"""
    if '-h' in args:
        return help    

    connectionInit(args)

    objectField = getParam(args,'-d')
    if objectField == None:
        print(help)
        return

    ofs = objectField.split(':')

    sObjectName = ofs[0]
    fieldName = ofs[1] if len(ofs) > 1 else None

    res = Sobjects.describe(sObjectName)
    if fieldName == None:
        print(simplejson.dumps(res['fields'], indent=4))
    else:
        sibbling = objectUtil.getSiblingWhere(res['fields'],'name',fieldName)['object']

        print(simplejson.dumps(sibbling, indent=4))

def option_l(args):
    if '-h' in args:
        help = """
        -l --> current org limits consuptions. """
        return help
    connectionInit(args)
    action = '/services/data/v51.0/limits'
    res = restClient.callAPI(action)
    records = []
    for key in res.keys():
        record = {
            'Limit':key,
            'Max':res[key]['Max'],
            'Remaining':res[key]['Remaining'],
        }
        record['Percent Remaining'] =  100 *(res[key]['Remaining']/res[key]['Max']) if res[key]['Max']>0 else 0
        record['__color__'] = ''

        if record['Max'] != 0:
            if record['Percent Remaining']<50:
                record['__color__'] = utils.CYELLOW
            if record['Percent Remaining']<25:
                record['__color__'] = utils.CYELLOW
        records.append(record)

    utils.printFormated(records)

def option_o(args):
    if '-h' in args:
        help = """
        -o --> List all Objects
            -like name
            objectName  --> get one row from the object
            objectName:Id --> get the row especified by the Id"""
        return help
        
    connectionInit(args)

    if '-name' in args:
        obj = getParam(args,'-name')
        chunks = obj.split(':')

        if len(chunks) == 1:
            q = f"select fields(all) from {chunks[0]} limit 1"
        else:
            q = f"select fields(all) from {chunks[0]} where Id='{chunks[1]}'"

        print('pre q:'+q)
        _queryAndPrint(q)
        return
        
    like = getParam(args,'-like') if '-like' in args else None    
    count = True if '-count' in args else False   

    objs = Sobjects.listObjects()
    if like is not None:
        if ':' not in like:
            like = f"name:{like}"        
        ls = like.split(':')
        outs = []
        for obj in objs:
      #      print(f"{str(ls[1])} . {str(obj[ls[0]])}")
            if str(ls[1]).lower() in str(obj[ls[0]]).lower():
                outs.append(obj) 
    else:
        outs = objs

    if count==True:
        for out in outs:
            if out['queryable'] == True:
                print("Quering objects row count.")
                print("", end=".")          
                try:
                    if out['name'] == 'AccountUserTerritory2View':
                        print()
                    c = query.query(f" select count(Id) from {out['name']}",raiseEx=True)
                    out['count'] = c['records'][0]['expr0']
                except Exception as e:
                    out['count'] = 'E'
            else:
                out['count'] = '-'
    
    #print(simplejson.dumps(objs, indent=4))
    utils.printFormated(outs,'name:label:associateParentEntity:associateParentEntity:queryable:count')

def option_default(args):
    help = """
        -default:set key value --> defaults the specified key-value
        -default:del key    --> deletes the default
        -default:get key    --> displays the current value
        """

    if '-h' in args:
        return help   

    if "-default:del" in args:
        key = getParam(args,'-default:del',1)
        if key == None:
            restClient.glog().info(f'Key not found in the provided arguments. {args}')
            return 
        restClient.delConfigVar(key)
        return

    if "-default:set" in args: 
        key = getParam(args,'-default:set',1)
        value = getParam(args,'-default:set',2)
        if key == None or value==None:
            print(help)
            return

        restClient.setConfigVar(key,value)
        print(f"Default value for {key} is {value}")

    if "-default:get" in args: 
        key = getParam(args,'-default:get',1)
        value =restClient.getConfigVar(key)
        print(f"Default value for {key} is {value}")

def option_history(args):
    if '-h' in args:
        help = """
        -history"""
        return help
    print()
    print('HISTORY:')
    vars = jsonFile.read(varsFile)
    for line in vars['history']:
        print(line)

def option_logs(args):
    help = """
        -logs:ls --> list last logs
            -limit how many, default 50 max 200
            -user "field:value", filter for specified user. The user can be specified by any field in the User Object. Examples:
                    "Id:XXXXXXXX"
                    "name:Onboarding Site Guest User"
                    "FirstName:Onboarding"
                    "ProfileId:00e3O000000IHneQAG"
        -logs --> parse logs
            -logs Id --> parse the logs with the provided Id
            -logs -last:X --> parses the last X logs. 
        """

    if '-h' in args:
        return help

    if '-logs:ls' in args:
        user = getParam(args,'-user') 
        last = getParam(args,'-last') 
        lim = 50 if last == None else last

        connectionInit(args)

        id = None
        if user != None:
            chunks = user.split(":")
            key = chunks[0]
            value = chunks[1]
            id = query.queryField(f"Select Id from User where {key}='{value}'") if key.lower()!='id' else value
        debugLogs.printLogs(logUserId=id,limit=lim)
    else:
        logId = getParam(args,'-logs')    
        last = getParam(args,'-last')    

        if logId == None and last == None:
            print(help)
            return

        connectionInit(args)
        debugLogs.parseLog(logId,lastN=last)

    return
    printLimits = False if '-nolimits' in args else True
    userDebug = False if '-userDebug' in args else True
    last = True if '-last' in args else False
    lastN = getParam(args,'-last')  if last == True else 1

    isLogUserId = True if '-userId' in args else False
    logUserId = getParam(args,'-userId')  if isLogUserId == True else None

    if last == True:
        logId = None
        
    if logId == 'latest':
        logId = None

    debugLogs.parseLog(logId,printLimits=printLimits,userDebug=userDebug,lastN=lastN,logUserId=logUserId)

def option_h(args):
    module = __import__('InCli')

    funcs = getmembers(sys.modules[__name__], isfunction)
    functions = [func[0] for func in funcs if func[0].startswith('option_')]

    for f in functions:
        if f == 'option_h':
            continue

        print(eval(f'{f}(args)'))      

def connectionInit(argsIn):

    userName_or_ofgAlias = getParam(argsIn,'-u') 
    restClient.init(userName_or_ofgAlias)

def main(argsIn):
    try:
    #    restClient.setConfigFile('/Users/uormaechea/Documents/Dev/python/Industries/config/ConnectionsParams.json')
    #    restClient.initWithConfig('DEVNOSCAT2')

        restClient.setLoggingLevel(logging.INFO)
      #  connectionInit(argsIn)

        # "uormaechea.devnoscat3@nos.pt"
        if '-h' in argsIn:
            help = """
        -u --> username or org alias to be used to log into the org. 
               Can be set to a default value --> -set u "xx"adas.com"
        """
            print(help)

        funcs = getmembers(sys.modules[__name__], isfunction)
        functions = [func[0] for func in funcs if func[0].startswith('option_')]

        args = []
        for argv in argsIn:
            if argv == '|':
                break
            args.append(argv)

        for arg in args:
            ar = arg.split(':')[0][1:]
            ar = f"option_{ar}"
            if ar in functions:
         #       module = __import__('InCli')
                eval(f'{ar}(args)')
          #      func = getattr(module, ar)
          #      func(args)

        if '-h' in argsIn:
            print()
            print("SFDX Commands:")
            print(" - sfdx force:org:list --verbose --all --> to list all authorized Orgs and Connection Status")
            print(" - sfdx auth:web:login -r 'Instance Url' -a 'Alias' --> to re-authorize")
            print(" - sfdx auth:web:login -u 'userName'  -a 'alias' --> to authorize and Org")
            print()

    except Exception as e:
        utils.printException(e)
        traceback.format_exc()

if __name__ == '__main__':
    main(sys.argv)

